#import <Foundation/NSObject.h>

@class NSString;
@class NSAttributedString;
@class SCKSourceLocation;
@class NSMutableArray;
@class NSMutableDictionary;
@class SCKIvar, SCKProperty;

/**
 * SCKProgramComponent is an abstract class representing properties of some
 * component of a program.  This includes classes, functions, methods, and so
 * on.
 *
 * Program components in SourceCodeKit can be obtained both by runtime
 * introspection and by parsing source code.  
 */
@interface SCKProgramComponent : NSObject
/**
 * The name of this program component.
 */
@property (nonatomic, retain) NSString *name;
/**
 * The location where the component is defined, if visible.  
 */
@property (nonatomic, retain) SCKSourceLocation *definition;
/**
 * The location where the component is declared, if visible.  
 */
// FIXME: We should allow multiple declaration locations.
@property (nonatomic, retain) SCKSourceLocation *declaration;
/**
 * Documentation associated with this object.  This may be generated by an IDE,
 * extracted from headers, or read from some external source.
 */
@property (nonatomic, retain) NSAttributedString *documentation;
/**
 * The parent of this component.  
 */
@property (nonatomic, unsafe_unretained) SCKProgramComponent *parent;
/**
 * Returns whether the program component has only been parsed as a forward 
 * declaration in the source code.
 */
@property (nonatomic, readonly, assign) BOOL isForwardDeclaration;
@end

/**
 * A program component with a type.
 */
@interface SCKTypedProgramComponent : SCKProgramComponent
/** Objective-C type encoding of the component. */
@property (retain, nonatomic) NSString *typeEncoding;
@end

@interface SCKBundle : SCKProgramComponent
/**
 * All of the public symbols exported by the bundle.
 */
@property (retain, nonatomic) NSMutableArray *classes;
@property (retain, nonatomic) NSMutableArray *functions;
@end

@interface SCKClass : SCKProgramComponent
@property (nonatomic, unsafe_unretained) SCKClass *superclass;
@property (nonatomic, readonly, retain) NSMutableArray *subclasses;
@property (nonatomic, readonly, retain) NSMutableDictionary *categories;
@property (nonatomic, readonly, retain) NSMutableDictionary *methods;
@property (nonatomic, readonly, retain) NSMutableArray *ivars;
@property (nonatomic, readonly, retain) NSMutableArray *properties;
- (SCKIvar*)ivarForName: (NSString *)name;
- (SCKProperty*)propertyForName: (NSString *)aProperty;
- (id) initWithClass: (Class)cls;
@end

@interface SCKProtocol : SCKProgramComponent
@property (nonatomic, readonly, retain) NSMutableDictionary *requiredMethods;
@property (nonatomic, readonly, retain) NSMutableDictionary *optionalMethods;
@property (nonatomic, readonly, retain) NSMutableArray *requiredProperties;
@property (nonatomic, readonly, retain) NSMutableArray *optionalProperties;
- (SCKProperty*)requiredPropertyForName: (NSString *)aProperty;
- (SCKProperty*)optionalPropertyForName: (NSString *)aProperty;
@end

@interface SCKCategory : SCKProgramComponent
@property (nonatomic, readonly, retain) NSMutableDictionary *methods;
@property (nonatomic, readonly, retain) NSMutableArray *properties;
- (SCKProperty*)propertyForName: (NSString *)aProperty;
@end

@interface SCKMethod : SCKTypedProgramComponent
@property (nonatomic, assign) BOOL isClassMethod;
@end

@interface SCKIvar : SCKTypedProgramComponent
@property (nonatomic, assign) BOOL isIBOutlet;
@end

@interface SCKFunction : SCKTypedProgramComponent
@end

@interface SCKGlobal : SCKTypedProgramComponent
@end

@interface SCKProperty : SCKTypedProgramComponent
@property (nonatomic, retain) NSString *attributes;
@property (nonatomic, assign) BOOL isIBOutlet;
@end

@interface SCKMacro : SCKTypedProgramComponent
@end

/**
 * Enumerated type value.  This encapsulates the name and value of the
 * enumeration value.  
 */
@interface SCKEnumerationValue : SCKProgramComponent
/**
 * Returns the value of the enumeration.
 */
@property long long longLongValue;
/**
 * The name of the enumeration.
 */
@property (nonatomic, retain) NSString *enumerationName;
@end

@interface SCKEnumeration : SCKTypedProgramComponent
/**
 * Array of SCKEnumerationValue instances that describe the values in this
 * enumeration.
 */
@property (nonatomic, retain) NSMutableDictionary *values;
@end
